# TODO create a functional programming interface for generating tasks. This includes
#   - Polygon sweeps, both for convex and non convex polygons
#   - Task generator for different categories: Roads, boundaries, buildings
#   - Environment generator, which can define a search area based on a polygon incasing the area.
#     The environment should take potental dangerous elements in the terrain into account:
#       * Powerlines, tall buildings, etc. (This should be configurable)


def generate_search_tasks(geometries):
    # normalize_coordinates(geometries)
    pass
